#include "../m_pd.h"
#include <time.h>

static t_class *pattern_class;

typedef struct _pattern {
	t_object x_obj;
	t_clock *x_clock;
	t_float x_f1, x_f2;
	double x_tempo, x_deltime;
	int x_hit, x_thym, x_pat, x_i, x_acrt, x_acnt;
	unsigned int x_state; // resets to 0 after passing 4,294,967,295
	t_outlet *b_out, *q_out, *p_out, *d_out, *i_out;
} t_pattern;

//{ #region random

int add_thym(void) {
	return (time(0) - 14400) % 86400; // resets to 0 at midnight
}

int timeseed(int thym) {
	static unsigned int pat_nextseed = 1489853723;
	pat_nextseed = pat_nextseed * thym + 938284287;
	return (pat_nextseed & 0x7fffffff);
}

int nextr(t_pattern *x, int n) {
	int nval;
	int range = (n < 1 ? 1 : n);
	x->x_state = x->x_state * 472940017 + 832416023;
	nval = ((double)range) * ((double)x->x_state) / 4294967296. + 1;
	return nval;
}

void pattern_seed(t_pattern *x, t_floatarg f) {
	if (f == 0) {
		x->x_thym = add_thym();
		x->x_state = x->x_thym;
	} else {
		x->x_thym = 0;
		x->x_state = f;
	}
}

void pattern_peek(t_pattern *x) {
	startpost("%u", x->x_state);
	if (x->x_thym != 0) post(" (t:%d)", x->x_thym);
	else endpost();
	post("%d beats at %f msecs", x->x_pat, x->x_deltime);
}

void pattern_accurate(t_pattern *x, t_floatarg f) {
	x->x_acrt = f;
}

//} #endregion random

//{ #region metro

void pattern_ft1(t_pattern *x, t_floatarg g) {
	x->x_tempo = (g <= 0 ? 1 : g);
	x->x_deltime = g / x->x_pat;
}

void pattern_tick(t_pattern *x) {
	x->x_hit = 0;
	outlet_float(x->i_out, x->x_i);
	if (x->x_i % (int)x->x_acnt == 0 && nextr(x, x->x_acrt) - 1)
		outlet_bang(x->b_out);
	else outlet_bang(x->q_out);
	x->x_i++;
	if (!x->x_hit)
		clock_delay(x->x_clock, x->x_deltime
		+ (x->x_pat == 1 ? 0.0001 : 0)); // prevents 1 beat from repeating
}

void pattern_float(t_pattern *x, t_float f) {
	if (f != 0) {
		x->x_i = 0;
		int n1 = nextr(x, x->x_f1);
		int n2 = nextr(x, x->x_f2);
		if (n1 == 1 || n2 == 1)
			x->x_acnt = nextr(x, (n1 != 1 ? n1 : n2));
		else x->x_acnt = (nextr(x, 2) - 1 ? n1 : n2);
		x->x_pat = n1 * n2;
		x->x_deltime = x->x_tempo / x->x_pat;
		outlet_float(x->p_out, x->x_pat);
		outlet_float(x->d_out, x->x_deltime);
		pattern_tick(x);
	}
	else clock_unset(x->x_clock);
	x->x_hit = 1;
}

void pattern_bang(t_pattern *x) {
	pattern_float(x, 1);
}

void pattern_stop(t_pattern *x) {
	pattern_float(x, 0);
}

void pattern_free(t_pattern *x) {
	clock_free(x->x_clock);
}

//} #endregion metro

void *pattern_new(t_floatarg tempo, t_floatarg f1, t_floatarg f2) {
	t_pattern *x = (t_pattern *)pd_new(pattern_class);
	
	x->x_pat = 1;
	x->x_acrt = 5; // 1 in 5 chance that an accent will be soft
	pattern_ft1(x, tempo);
	x->x_hit = 0;
	x->x_clock = clock_new(x, (t_method)pattern_tick);
	
	x->x_f1 = (f1 > 1 ? f1 : 4);
	x->x_f2 = (f2 > 1 ? f2 : 1);
	x->x_thym = add_thym();
	x->x_state = timeseed(x->x_thym);
	
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("ft1"));
	floatinlet_new(&x->x_obj, &x->x_f1);
	floatinlet_new(&x->x_obj, &x->x_f2);
	
	x->b_out = outlet_new(&x->x_obj, &s_bang);  // accented beats
	x->q_out = outlet_new(&x->x_obj, &s_bang);  // every other beat
	x->d_out = outlet_new(&x->x_obj, &s_float); // delay after dividing the tempo
	x->p_out = outlet_new(&x->x_obj, &s_float); // denominator
	x->i_out = outlet_new(&x->x_obj, &s_float); // counter
	return (x);
}

void pattern_setup(void) {
	pattern_class = class_new(gensym("pattern"),
		(t_newmethod)pattern_new, (t_method)pattern_free,
		sizeof(t_pattern), CLASS_DEFAULT,
		A_DEFFLOAT, A_DEFFLOAT, A_DEFFLOAT, 0);
	
	class_addbang(pattern_class, pattern_bang);
	class_addmethod(pattern_class, (t_method)pattern_seed,
		gensym("seed"), A_DEFFLOAT, 0);
	class_addmethod(pattern_class, (t_method)pattern_peek,
		gensym("peek"), 0);
	class_addmethod(pattern_class, (t_method)pattern_stop, gensym("stop"), 0);
	class_addmethod(pattern_class, (t_method)pattern_ft1, gensym("ft1"),
		A_FLOAT, 0);
	class_addmethod(pattern_class, (t_method)pattern_accurate,
		gensym("acrt"), A_FLOAT, 0);
	class_addfloat(pattern_class, (t_method)pattern_float);
}

